[![FactSet](https://raw.githubusercontent.com/factset/enterprise-sdk/main/docs/images/factset-logo.svg)](https://www.factset.com)

# Formula client library for Python

[![API Version](https://img.shields.io/badge/api-v1.8.0-blue)](https://developer.factset.com/api-catalog/formula-api)
[![PyPi](https://img.shields.io/pypi/v/fds.sdk.Formula)](https://pypi.org/project/fds.sdk.Formula/)
[![Apache-2 license](https://img.shields.io/badge/license-Apache2-brightgreen.svg)](https://www.apache.org/licenses/LICENSE-2.0)


**FactSet Formula API**

FactSet’s Formula API is a modern, flexible, formula-based API that enables users to access FactSet’s wide range of financial data and content. The API offers two endpoints, one optimized for time-series analysis and one designed for cross-sectional analysis, providing users a simplified interface into FactSet’s expansive offering. By providing two endpoints, it allows for the optimization of user workflows, while reducing complexity.

Leverage the power and flexibility of the Formula API to -
  * Pull data from across most content sets that a user has access to in a single request
  * Include business logic and mathematical operations in request
  * Submit a dynamic universe in both endpoints
  * Return the fsymId to easily combine with other FactSet content / products
  * Set the trading calendar
  * Define custom display names

**Formula API Request Builder**

The Formula API Request Builder provides users everything they need to form a Formula API request. In the Request Builder, you can select identifiers, build a universe expression, select FQL or Screening formulas, easily apply business logic and mathematical functions to the FQL or Screening formulas, specify optional parameters, and construct a GET or POST request. The Request Builder eliminates the need to have previous FQL and Screening knowledge and allows you to quickly find your desired data items and form the request.

The Formula API Request Builder can be accessed by navigating to [https://developer.factset.com/formula-api-request-builder](https://developer.factset.com/formula-api-request-builder) and logging in using your FactSet.net ID. When using the Request Builder to construct requests for the one of the Formula API's endpoints, be sure to toggle to the correct endpoint at the top of the page.

**How to Check the Health and Availability of the Formula API**

Please use the below endpoint to check the health and availability of the Formula API. You must be authorized for this API to use the Health endpoint.

[https://api.factset.com/formula-api/health](https://api.factset.com/formula-api/health)

**How to Programmatically Download API Specification File**

You can download the FactSet Formula API Specification File in .yaml. using the \"Download Spec\" button to the right of the version number. This specification can then be used for Codegen to create your own SDKs.


This Python package is automatically generated by the [OpenAPI Generator](https://openapi-generator.tech) project:

- API version: 1.8.0
- SDK version: 3.0.0
- Build package: org.openapitools.codegen.languages.PythonClientCodegen

For more information, please visit [https://developer.factset.com/contact](https://developer.factset.com/contact)

## Requirements

* Python >= 3.7

## Installation

### Poetry

```shell
poetry add fds.sdk.utils fds.sdk.Formula==3.0.0
```

### pip

```shell
pip install fds.sdk.utils fds.sdk.Formula==3.0.0
```

## Usage

1. [Generate authentication credentials](../../../../README.md#authentication).
2. Setup Python environment.
   1. Install and activate python 3.7+. If you're using [pyenv](https://github.com/pyenv/pyenv):

      ```sh
      pyenv install 3.9.7
      pyenv shell 3.9.7
      ```

   2. (optional) [Install poetry](https://python-poetry.org/docs/#installation).
3. [Install dependencies](#installation).
4. Run the following:

> [!IMPORTANT]
> The parameter variables defined below are just examples and may potentially contain non valid values. Please replace them with valid values.

### Example Code

```python
from fds.sdk.utils.authentication import ConfidentialClient

import fds.sdk.Formula
from fds.sdk.Formula.api import batch_processing_api
from fds.sdk.Formula.models import *
from dateutil.parser import parse as dateutil_parser
from pprint import pprint

# See configuration.py for a list of all supported configuration parameters.

# Examples for each supported authentication method are below,
# choose one that satisfies your use case.

# (Preferred) OAuth 2.0: FactSetOAuth2
# See https://github.com/FactSet/enterprise-sdk#oauth-20
# for information on how to create the app-config.json file
#
# The confidential client instance should be reused in production environments.
# See https://github.com/FactSet/enterprise-sdk-utils-python#authentication
# for more information on using the ConfidentialClient class
configuration = fds.sdk.Formula.Configuration(
    fds_oauth_client=ConfidentialClient('/path/to/app-config.json')
)

# Basic authentication: FactSetApiKey
# See https://github.com/FactSet/enterprise-sdk#api-key
# for information how to create an API key
# configuration = fds.sdk.Formula.Configuration(
#     username='USERNAME-SERIAL',
#     password='API-KEY'
# )

# Enter a context with an instance of the API client
with fds.sdk.Formula.ApiClient(configuration) as api_client:
    # Create an instance of the API class
    api_instance = batch_processing_api.BatchProcessingApi(api_client)
    id = "id_example" # str | Batch Request identifier.

    try:
        # Returns the response for a Batch Request
        # example passing only required values which don't have defaults set
        api_response_wrapper = api_instance.get_batch_data(id)

        # This endpoint returns a response wrapper that contains different types of responses depending on the query.
        # To access the correct response type, you need to perform one additional step, as shown below.
        if api_response_wrapper.get_status_code() == 200:
            api_response = api_response_wrapper.get_response_200()
        if api_response_wrapper.get_status_code() == 202:
            api_response = api_response_wrapper.get_response_202()

        pprint(api_response)
    except fds.sdk.Formula.ApiException as e:
        print("Exception when calling BatchProcessingApi->get_batch_data: %s\n" % e)

    # # Get response, http status code and response headers
    # try:
    #     # Returns the response for a Batch Request
    #     api_response_wrapper, http_status_code, response_headers = api_instance.get_batch_data_with_http_info(id)

    #     # This endpoint returns a response wrapper that contains different types of responses depending on the query.
    #     # To access the correct response type, you need to perform one additional step, as shown below.
    #     if api_response_wrapper.get_status_code() == 200:
    #         api_response = api_response_wrapper.get_response_200()
    #     if api_response_wrapper.get_status_code() == 202:
    #         api_response = api_response_wrapper.get_response_202()

    #     pprint(api_response)
    #     pprint(http_status_code)
    #     pprint(response_headers)
    # except fds.sdk.Formula.ApiException as e:
    #     print("Exception when calling BatchProcessingApi->get_batch_data: %s\n" % e)

    # # Get response asynchronous
    # try:
    #     # Returns the response for a Batch Request
    #     async_result = api_instance.get_batch_data_async(id)
    #     api_response_wrapper = async_result.get()

    #     # This endpoint returns a response wrapper that contains different types of responses depending on the query.
    #     # To access the correct response type, you need to perform one additional step, as shown below.
    #     if api_response_wrapper.get_status_code() == 200:
    #         api_response = api_response_wrapper.get_response_200()
    #     if api_response_wrapper.get_status_code() == 202:
    #         api_response = api_response_wrapper.get_response_202()

    #     pprint(api_response)
    # except fds.sdk.Formula.ApiException as e:
    #     print("Exception when calling BatchProcessingApi->get_batch_data: %s\n" % e)

    # # Get response, http status code and response headers asynchronous
    # try:
    #     # Returns the response for a Batch Request
    #     async_result = api_instance.get_batch_data_with_http_info_async(id)
    #     api_response_wrapper, http_status_code, response_headers = async_result.get()

    #     # This endpoint returns a response wrapper that contains different types of responses depending on the query.
    #     # To access the correct response type, you need to perform one additional step, as shown below.
    #     if api_response_wrapper.get_status_code() == 200:
    #         api_response = api_response_wrapper.get_response_200()
    #     if api_response_wrapper.get_status_code() == 202:
    #         api_response = api_response_wrapper.get_response_202()

    #     pprint(api_response)
    #     pprint(http_status_code)
    #     pprint(response_headers)
    # except fds.sdk.Formula.ApiException as e:
    #     print("Exception when calling BatchProcessingApi->get_batch_data: %s\n" % e)

```

### Using Pandas

To convert an API response to a Pandas DataFrame, it is necessary to transform it first to a dictionary.
```python
import pandas as pd

response_dict = api_response.to_dict()['data']

simple_json_response = pd.DataFrame(response_dict)
nested_json_response = pd.json_normalize(response_dict)
```

### Debugging

The SDK uses the standard library [`logging`](https://docs.python.org/3/library/logging.html#module-logging) module.

Setting `debug` to `True` on an instance of the `Configuration` class sets the log-level of related packages to `DEBUG`
and enables additional logging in Pythons [HTTP Client](https://docs.python.org/3/library/http.client.html).

**Note**: This prints out sensitive information (e.g. the full request and response). Use with care.

```python
import logging
import fds.sdk.Formula

logging.basicConfig(level=logging.DEBUG)

configuration = fds.sdk.Formula.Configuration(...)
configuration.debug = True
```

### Configure a Proxy

You can pass proxy settings to the Configuration class:

* `proxy`: The URL of the proxy to use.
* `proxy_headers`: a dictionary to pass additional headers to the proxy (e.g. `Proxy-Authorization`).

```python
import fds.sdk.Formula

configuration = fds.sdk.Formula.Configuration(
    # ...
    proxy="http://secret:password@localhost:5050",
    proxy_headers={
        "Custom-Proxy-Header": "Custom-Proxy-Header-Value"
    }
)
```

### Custom SSL Certificate

TLS/SSL certificate verification can be configured with the following Configuration parameters:

* `ssl_ca_cert`: a path to the certificate to use for verification in `PEM` format.
* `verify_ssl`: setting this to `False` disables the verification of certificates.
  Disabling the verification is not recommended, but it might be useful during
  local development or testing.

```python
import fds.sdk.Formula

configuration = fds.sdk.Formula.Configuration(
    # ...
    ssl_ca_cert='/path/to/ca.pem'
)
```

### Request Retries

In case the request retry behaviour should be customized, it is possible to pass a `urllib3.Retry` object to the `retry` property of the Configuration.

```python
from urllib3 import Retry
import fds.sdk.Formula

configuration = fds.sdk.Formula.Configuration(
    # ...
)

configuration.retries = Retry(total=3, status_forcelist=[500, 502, 503, 504])
```


## Documentation for API Endpoints

All URIs are relative to *https://api.factset.com/formula-api/v1*

Class | Method | HTTP request | Description
------------ | ------------- | ------------- | -------------
*BatchProcessingApi* | [**get_batch_data**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchProcessingApi.md#get_batch_data) | **GET** /batch-result | Returns the response for a Batch Request
*BatchProcessingApi* | [**get_batch_data_with_post**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchProcessingApi.md#get_batch_data_with_post) | **POST** /batch-result | Returns the status for a Batch Request
*BatchProcessingApi* | [**get_batch_status**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchProcessingApi.md#get_batch_status) | **GET** /batch-status | Returns the status for a Batch Request
*BatchProcessingApi* | [**get_batch_status_with_post**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchProcessingApi.md#get_batch_status_with_post) | **POST** /batch-status | Returns the status for a Batch Request
*CrossSectionalApi* | [**get_cross_sectional_data**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalApi.md#get_cross_sectional_data) | **GET** /cross-sectional | Retrieve data items (Screening formulas) for a list of identifiers or defined universe.
*CrossSectionalApi* | [**get_cross_sectional_data_for_list**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalApi.md#get_cross_sectional_data_for_list) | **POST** /cross-sectional | Retrieve data items (Screening formulas) for a list of identifiers or defined universe.
*TimeSeriesApi* | [**get_time_series_data**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesApi.md#get_time_series_data) | **GET** /time-series | Retrieve data items (FQL formulas) for a list of identifiers or defined universe.
*TimeSeriesApi* | [**get_time_series_data_for_list**](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesApi.md#get_time_series_data_for_list) | **POST** /time-series | Retrieve data items (FQL formulas) for a list of identifiers or defined universe.


## Documentation For Models

 - [BatchData](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchData.md)
 - [BatchDataRequest](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchDataRequest.md)
 - [BatchDataRequestData](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchDataRequestData.md)
 - [BatchDataResponse](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchDataResponse.md)
 - [BatchStatus](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchStatus.md)
 - [BatchStatusResponse](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/BatchStatusResponse.md)
 - [CrossSectionalRequest](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalRequest.md)
 - [CrossSectionalRequestData](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalRequestData.md)
 - [CrossSectionalResponse](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalResponse.md)
 - [CrossSectionalResponseObjectItems](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalResponseObjectItems.md)
 - [CrossSectionalResultObjectFlattened](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalResultObjectFlattened.md)
 - [CrossSectionalResultObjectFlattenedAddl](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalResultObjectFlattenedAddl.md)
 - [CrossSectionalResultObjectNonflattened](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalResultObjectNonflattened.md)
 - [CrossSectionalResultObjectNonflattenedResultAttribute](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/CrossSectionalResultObjectNonflattenedResultAttribute.md)
 - [ErrorDetail](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/ErrorDetail.md)
 - [ErrorObject](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/ErrorObject.md)
 - [TimeSeriesRequest](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesRequest.md)
 - [TimeSeriesRequestData](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesRequestData.md)
 - [TimeSeriesResponse](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResponse.md)
 - [TimeSeriesResponseObjectItems](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResponseObjectItems.md)
 - [TimeSeriesResultObjectFlattened](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectFlattened.md)
 - [TimeSeriesResultObjectFlattenedAddl](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectFlattenedAddl.md)
 - [TimeSeriesResultObjectNonflattened](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectNonflattened.md)
 - [TimeSeriesResultObjectNonflattenedArray](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectNonflattenedArray.md)
 - [TimeSeriesResultObjectNonflattenedBase](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectNonflattenedBase.md)
 - [TimeSeriesResultObjectNonflattenedMatrix](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectNonflattenedMatrix.md)
 - [TimeSeriesResultObjectNonflattenedScalar](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectNonflattenedScalar.md)
 - [TimeSeriesResultObjectNonflattenedTimeseriesObject](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesResultObjectNonflattenedTimeseriesObject.md)
 - [TimeSeriesScalarValue](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesScalarValue.md)
 - [TimeSeriesTimeseriesObject](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/TimeSeriesTimeseriesObject.md)
 - [WarningsObject](https://github.com/FactSet/enterprise-sdk/tree/main/code/python/Formula/v1/docs/WarningsObject.md)


## Documentation For Authorization


## FactSetApiKey

- **Type**: HTTP basic authentication


## FactSetOAuth2

- **Type**: OAuth
- **Flow**: application
- **Authorization URL**: 
- **Scopes**: N/A


## Notes for Large OpenAPI documents
If the OpenAPI document is large, imports in fds.sdk.Formula.apis and fds.sdk.Formula.models may fail with a
RecursionError indicating the maximum recursion limit has been exceeded. In that case, there are a couple of solutions:

Solution 1:
Use specific imports for apis and models like:
- `from fds.sdk.Formula.api.default_api import DefaultApi`
- `from fds.sdk.Formula.model.pet import Pet`

Solution 2:
Before importing the package, adjust the maximum recursion limit as shown below:
```
import sys
sys.setrecursionlimit(1500)
import fds.sdk.Formula
from fds.sdk.Formula.apis import *
from fds.sdk.Formula.models import *
```

## Contributing

Please refer to the [contributing guide](../../../../CONTRIBUTING.md).

## Copyright

Copyright 2022 FactSet Research Systems Inc

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

